# -*- coding: utf-8 -*-
"""
Kanji.py
手書き漢字
(1) 国文学研究資料館: 3832文字, 140424個, 64x64ピクセル
http://codh.rois.ac.jp/kmnist/
https://github.com/rois-codh/kmnist
(2) 東京大学史料編纂所: 5896文字, 325261個, 大きさ不定
https://lab.hi.u-tokyo.ac.jp/datasets/kuzushiji
"""

import os
import numpy as np
import torch
from torchvision import datasets, transforms
import time
import utils

# DATA = 'KKanji'/'TKanji'
def dataset(DATA, MODEL, in_size):

    assert (in_size[1] == in_size[2])

    fn_image = '%s_images_%d.npz' % (DATA, in_size[1])
    fn_label = '%s_labels.npz' % DATA
    path_image = os.path.join(os.getcwd(), DATA, fn_image)
    path_label = os.path.join(os.getcwd(), DATA, fn_label)

    # npzファイルがないときはdatasetを読み込んで出力する
    # 続けて行うとキャッシュに入っているので数分の1の時間ですむ
    if (not os.path.isfile(path_image)) or \
       (not os.path.isfile(path_label)):
        # 前処理
        transform = transforms.Compose([
            transforms.ToTensor(),  # 必須
            transforms.Resize((in_size[1], in_size[2])),  # リサイズ
            # gray=RGB平均
            transforms.Lambda(lambda x: (x[0, :, :] + x[1, :, :] + x[2, :, :]).view(1, in_size[1], in_size[2]) / 3.0),
        ])

        # ImageFolder : 画像とラベル(=ディレクトリの通し番号=0,1,2,...)のペアを返す
        # 全データセットを読み込む
        imagedir = 'kkanji2' if DATA == 'KKanji' else 'tkanji'
        root = os.path.join(os.getcwd(), DATA, imagedir)
        all_set = datasets.ImageFolder(root, transform=transform)
        #print(len(all_set))

        # 配列作成
        images = np.zeros((len(all_set), in_size[1], in_size[2]), dtype='u1')
        labels = np.zeros(len(all_set), dtype='i8')  # i4は不可

        # dataset作成(numpy型)
        t0 = time.time()
        for i, (image, label) in enumerate(all_set):
            image = image.numpy() * 255.0
            images[i] = image.astype('u1').copy()
            labels[i] = label
            if i % 10000 == 0:  # 経過表示
                t1 = time.time(); print('%d %.2f' % (i, t1 - t0)); t0 = t1

        # save
        np.savez(path_image, images)
        np.savez(path_label, labels)
        print('Save %s, %s' % (fn_image, fn_label))

    # npzファイルを読み込む
    d_image = np.load(path_image)['arr_0']  # u1
    d_label = np.load(path_label)['arr_0']  # i8
    print('Load %s, %s' % (fn_image, fn_label))

    #print(d_image[0].shape)  # (64, 64)
    #plt.imshow(d_image[0], cmap='gray')

    assert len(d_image) == len(d_label)

    # 訓練データとテストデータに分ける
    train_ratio = 6 / 7
    rand = torch.rand(len(d_image))
    train_set, test_set = [], []
    for i in range(len(d_image)):
        # numpy(0-255, u1) -> tensor(0-1, f4)
        t_image = torch.tensor(d_image[i]).float() / 255.0
        # 色変換, 白地に黒字
        if DATA == 'KKanji':
            # 白黒反転
            t_image = 1 - t_image
        elif DATA == 'TKanji':
            pass  # 何もしないのがベスト
            # 平均を境にした白黒2値
            #thres = t_image.mean() - 0.05
            #thres = t_image.mean() * 0.9
            #t_image[t_image >  thres] = 1
            #t_image[t_image <= thres] = 0
        # 配列を変換してコピー
        if (MODEL == 0) or (MODEL == 1):
            # grayのままコピー
            image = t_image.view(1, in_size[1], in_size[2])
        elif (MODEL == 2) or (MODEL == 3):
            # gray -> R=G=Bの3色
            image = torch.zeros((3, in_size[1], in_size[2]))
            image[0, :, :] = image[1, :, :] = image[2, :, :] = t_image[:, :]
        # 乱数で分配
        if rand[i] < train_ratio:
            train_set.append([image, d_label[i]])
        else:
            test_set.append([image, d_label[i]])

    # テスト用データセットのコピー
    test0_set = test_set.copy()

    # debug
    #_plot_image(train_set)

    """
    # debug
    num = 0
    #for sdir in os.listdir('./TKanji/tkanji'):
    for sdir in os.listdir('./KKanji/kkanji2'):
        if num % 50 == 0:
            print('%4d ' % num, end='')
        print(utils.unicode2char(sdir), end='')
        num += 1
        if num % 50 == 0:
            print('')
    """

    return train_set, test_set, test0_set

"""
# debug
import matplotlib.pyplot as plt
def _plot_image(dataset):
    plt.figure('Kanji dataset', figsize=(9, 6))
    nrow = 12
    ncol = 20
    dpos = 20000
    dstep = 300
    ipos = 0
    for i in range(nrow):
        for j in range(ncol):
            #pos = (i * ncol) + j
            image, label = dataset[dpos]
            #plt.imshow(img.permute(1, 2, 0), cmap='gray')
            ax = plt.subplot(nrow, ncol, ipos + 1)
            plt.imshow(np.transpose(image, (1, 2, 0)), cmap='gray')
            #plt.imshow(image2)
            ax.set_axis_off()
            ax.set_title(label, fontsize=8, pad=0)
            dpos += dstep
            ipos += 1
    plt.show()
"""