# -*- coding: utf-8 -*-
"""
fit.py
"""

import time
import numpy as np
import torch
from torch.amp import autocast, GradScaler

# 訓練
def _train(model, device, train_loader, optimizer, criterion):
    model.train()
    for data, target in train_loader:
        data = data.to(device)
        target = target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = criterion(output, target)
        loss.backward()
        optimizer.step()

# テスト
def _test(model, device, test_loader, criterion):
    model.eval()
    test_loss = 0
    test_correct = 0
    labels = np.zeros(0, 'i4')
    with torch.no_grad():
        for data, target in test_loader:
            data = data.to(device)
            target = target.to(device)
            output = model(data)
            loss = criterion(output, target)
            test_loss += loss.item() * len(data)  # sum up batch loss
            pred = output.argmax(dim=1, keepdim=True)  # get the index
            test_correct += pred.eq(target.view_as(pred)).sum().item()
            labels = np.append(labels, pred.to("cpu").numpy())

    return test_loss / len(test_loader.dataset), test_correct, labels

# 学習
def fit(model, device, num_epochs, train_loader, test_loader, optimizer, criterion, modelfile, mixed=False):

    # 開始時刻
    t0 = time.time()
    t1 = t0

    # GPU転送
    model = model.to(device)

    # 損失と正解率を保存する配列
    history = np.zeros((0, 2))

    # log
    fp = open('CNN.log', 'wt', encoding='utf-8')
    max_acc = 0

    # エポックに関するループ
    for epoch in range(num_epochs + 1):

        # 訓練
        if epoch > 0:  # 初回は行わない
            if not mixed:
                _train(model, device, train_loader, optimizer, criterion)
            else:
                _train_mixed(model, device, train_loader, optimizer, criterion)

        # テスト
        test_loss, test_correct, test_labels = _test(model, device, test_loader, criterion)

        # テストの正解率
        test_acc = 100 * test_correct / len(test_loader.dataset) if len(test_loader.dataset) > 0 else 0 # 正解率

        # 損失と正解率を出力する
        t2 = time.time()
        print('%4d %.5f%7.3f (%d/%d)%8.1f(%5.1f)[sec]'
              % (epoch, test_loss, test_acc, test_correct, len(test_loader.dataset), t2 - t0, t2 - t1),
              flush=True)
        fp.write('%4d %.3f\n' % (epoch, test_acc))
        fp.flush()
        t1 = t2

        # 損失と正解率を配列に代入する
        item = np.array([test_loss, test_acc])
        history = np.vstack((history, item))

        # 正解率を更新した: パラメーター保存(restart+mojila用)
        if test_acc > max_acc:
            max_acc = test_acc
            torch.save(model.state_dict(), modelfile)

    print('max accuracy %.3f' % max_acc)
    print('Save %s' % modelfile)
    fp.close()

    return history, test_labels

# 訓練、テンソルコア用混合精度、結果は少し変わる
# 画像サイズが大きいとき(64以上)約2倍速くなる
# https://pytorch.org/docs/stable/notes/amp_examples.html
def _train_mixed(model, device, train_loader, optimizer, criterion):

    scaler = GradScaler('cuda')

    model.train()
    for data, target in train_loader:
        # これでもよい
        #data = data.to(device)
        #target = target.to(device)
        
        # cuda
        data = data.cuda()
        target = target.cuda()

        # 勾配初期化
        optimizer.zero_grad()

        # Runs the forward pass with autocasting.
        with autocast(device_type='cuda', dtype=torch.float16):   # モデルの演算精度を自動選択する
            output = model(data)
            loss = criterion(output, target)

        # 損失計算のスケール化、逆誤差伝搬
        # Scales loss.  Calls backward() on scaled loss to create scaled gradients.
        # Backward passes under autocast are not recommended.
        # Backward ops run in the same dtype autocast chose for corresponding forward ops.
        scaler.scale(loss).backward()  # スケール化されたlossのbackward関数を使用する

        # オプティマイザのstep処理
        # scaler.step() first unscales the gradients of the optimizer's assigned params.
        # If these gradients do not contain infs or NaNs, optimizer.step() is then called,
        # otherwise, optimizer.step() is skipped.
        scaler.step(optimizer)  # optimizer.step()の代わりに、scaler.step()を使用する

        # Scalerの更新
        # Updates the scale for next iteration.
        scaler.update()

        #output = model(data)
        #loss = criterion(output, target)
        #loss.backward()
        #optimizer.step()
