# -*- coding: utf-8 -*-
"""
utils.py
共通関数(編集不要)
"""

import os
import numpy as np
import matplotlib.pyplot as plt
import torch
from torch.utils.data import DataLoader

# dataloader
def dataloader(train_set, test_set, batch_size=50):
    train_loader, test_loader = None, None
    if train_set is not None:
        train_loader = DataLoader(train_set, batch_size=batch_size, shuffle=True)
    if test_set is not None:
        test_loader  = DataLoader(test_set,  batch_size=batch_size, shuffle=False)

    return train_loader, test_loader

# 乱数初期化
def torch_seed(seed = 123):
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.use_deterministic_algorithms = True

# モデルファイルを読み込む
def load_model(model, modelfile, device):
    if (modelfile is not None) and os.path.isfile(modelfile):
        model.load_state_dict(torch.load(modelfile, map_location=device, weights_only=True))
        print("Load %s" % modelfile)

"""
# モデルファイルを保存する
def save_model(net, DATA):
    modelfile = '%s.pth' % DATA
    torch.save(net.state_dict(), modelfile)
    print("Save %s" % modelfile)
"""

# Unicodeを文字に変換する
# U+4E00 -> 一
def unicode2char(unicode):
    u = '0x' + unicode[2:]
    return chr(int(u, 16))

# 学習曲線図形表示
def show_history(history):
    plt.figure('history', figsize=(9.5, 6.5))
    plt.plot(history[:, 0], label='test loss')
    plt.plot(history[:, 1], label='test accuracy')
    plt.grid(True)
    plt.legend()
    plt.show()

# 画像図形表示
def show_images_labels(strtitle, dataset, labels, strclasses, num_data,
    w=9.5, h=6.5, fontname=None, fontsize=8, nrow=10, ncol=20, chars=-1,
    incorrect=False, start=0, step=1, thres=0, cmap=None, transpose=False):

    # strtitle : ウィンドウのタイトル
    # dataset : データセット(画像,ラベル)
    # labels : ラベルのテスト結果
    # strclasses : ラベルの文字列
    # num_data : 各クラスのデータ数
    # incorrect=False : 正誤にかかわらずすべてのデータを図形表示する
    # incorrect=True : 間違ったデータのみ図形表示する
    # start : 図形表示する最初のデータ番号
    # step : 図形表示するデータ番号の間隔
    # thres : データ数がこれ以上の画像のみ表示する
    # cmap : カラーマップ(Noneのときはそのままの色)
    # transpose : 行と列を交換するか

    # 図形表示開始
    plt.figure(strtitle, figsize=(w, h))

    # フォント(日本語表示用)
    if fontname is not None:
        plt.rcParams['font.family'] = fontname

    num = 0
    for idx in range(start, len(dataset), step):
        # 画像とラベルを取得
        image, label = dataset[idx]
        
        # データ数に閾値を設ける
        if num_data[0, label] >= thres:

            # 色を1番目から3番目に移動する、必要なら行と列を転置する
            if transpose:
                image2 = np.transpose(image, (2, 1, 0))
            else:
                image2 = np.transpose(image, (1, 2, 0))
    
            # 条件に合うものを図形表示する
            if (incorrect == False) or (label != labels[idx]):
                # レイアウト
                ax = plt.subplot(nrow, ncol, num + 1)
                
                # 画像表示
                plt.imshow(image2, cmap=cmap)
                    
                # 軸非表示
                ax.set_axis_off()
    
                # タイトル
                if label != labels[idx]:
                    # 間違ったときは上に正解、下に不正解を表示する(2行, 赤)
                    title = '%s\n%s' % (strclasses[label][:chars], strclasses[labels[idx]][:chars])
                    color = 'r'
                else:
                    # 正解のときは正解を表示する(1行, 黒)
                    title = '%s' % strclasses[label][:chars]
                    color = 'k'
                ax.set_title(title, color=color, fontsize=fontsize, pad=0)
                
                # 図形番号++
                num += 1
                if (num >= nrow * ncol):
                    #print(start, idx) # debug
                    break

    plt.show()

# クラス別のデータ数と正解数(テストデータ)
def num_data_class(test_set, test_labels, strclasses):

    nclass = len(strclasses)
    num_data = np.zeros((2, nclass), 'i4')

    for idx in range(len(test_set)):
        image, label0 = test_set[idx]  # label0:正解ラベル
        label1 = test_labels[idx]      # label1:回答ラベル
        assert(label0 >= 0 and label0 < nclass and label1 >= 0 and label1 < nclass)
        num_data[0, label0] += 1
        num_data[1, label0] += 1 if (label0 == label1) else 0

    return num_data

# 正解率を出力する
def output_accuracy(test_set, test_labels, strclasses, num_data, thres):

    nclass = len(strclasses)

    # 全データの正解率(主にdebug用)
    if thres < 0:
        for iclass in range(nclass):
            ndata0 = num_data[0, iclass]
            ndata1 = num_data[1, iclass]
            accuracy = (ndata1 / ndata0) * 100 if (ndata0 > 0) else 0
            print('%d %s %.3f (%d/%d)' % (iclass, strclasses[iclass], accuracy, ndata1, ndata0))
    
    # データ数が少ない文字を除外した正解率
    else:
        nsum0, nsum1, mclass = 0, 0, 0
        for iclass in range(nclass):
            ndata0 = num_data[0, iclass]
            ndata1 = num_data[1, iclass]
            if ndata0 > thres:  # 閾値
               nsum0 += ndata0
               nsum1 += ndata1
               mclass += 1
        accuracy = (nsum1 / nsum0) * 100 if (nsum0 > 0) else 0
        print('accuracy = %.3f (%d/%d) (over %d class = %d/%d)' % (accuracy, nsum1, nsum0, thres, mclass, nclass))

# 文字ごとの統計をHTMLで出力する(KKanji用)
def output_class_html(strclasses, num_data, step):

    with open('table.txt', 'wt', encoding='utf-8') as f:
        f.write('<table border=\"1\">\n')
        for iclass in range(len(strclasses)):
            if iclass % step == 0:
                f.write('<tr><td>%d</td>' % iclass)
            f.write('<td>%s %d %d</td>' % (strclasses[iclass], num_data[0, iclass], num_data[1, iclass]))
            if (iclass % step == step - 1) or (iclass == len(strclasses) - 1):
                f.write('</tr>\n')
        f.write('</table>\n')

# debug : テストデータセットの確認
def plot_dataset(dataset, strclasses, w=9.5, h=7.5, nrow=20, ncol=20,
    startlabel=0, stoplabel=10, fontsize=8, color=False, transpose=False):

    # 図形表示開始
    plt.figure('sample of dataset', figsize=(w, h))

    # フォント(日本語表示用)
    plt.rcParams['font.family'] = 'MS gothic'

    icol = np.zeros(nrow, 'i4')
    for idx, (image, label) in enumerate(dataset):
        irow = label - startlabel
        if (irow >= 0) and (irow <= stoplabel - startlabel) and (icol[irow] < ncol):
            # 描画位置
            ipos = (irow * ncol) + icol[irow] + 1
            ax = plt.subplot(nrow, ncol, ipos)
    
            # 色を1番目から3番目に移動する、必要なら行と列を転置する
            if transpose:
                image2 = np.transpose(image, (2, 1, 0))
            else:
                image2 = np.transpose(image, (1, 2, 0))
    
            # 画像表示
            cmap = None if color else 'gray_r'
            plt.imshow(image2, cmap=cmap)
    
            # 軸非表示
            ax.set_axis_off()

            # ラベル表示(最初のカラムのみ)
            if (icol[irow] == 0) and (fontsize > 0):
                ax.set_title(strclasses[label], fontsize=fontsize, pad=0)

            # カラム位置++
            icol[irow] += 1

# debug : テストデータセットの確認(MNIST系)
def plot_MNIST(DATA, dataset, strclasses, label0, label1, ncol, w=9, h=6, name=True, transpose=False):
    label0 = min(label0, len(strclasses))
    label1 = min(label1, len(strclasses))
    
    plt.figure('%s dataset' % DATA, figsize=(w, h))
    plt.rcParams['font.family'] = 'MS gothic'

    ndone = np.zeros(len(strclasses), 'i4')
    for (image, label) in dataset:
        assert label >= 0 and label < len(strclasses)
        if ndone[label] < ncol and label >= label0 and label < label1:
            ipos = ncol * (label - label0) + ndone[label] + 1
            ax = plt.subplot(label1 - label0, ncol, ipos)
            if transpose:
                image2 = np.transpose(image, (2, 1, 0))
            else:
                image2 = np.transpose(image, (1, 2, 0))
            plt.imshow(image2, cmap='gray_r')
            ax.set_axis_off()
            if name and (ndone[label] == 0):
                ax.set_title(strclasses[label], fontsize=8, pad=0)
            ndone[label] += 1

# データセットの平均と標準偏差を計算する（正規化用）
# ncolor = 1/3 : gray, color
def mean_std(dataset, ncolor):
    depth, rows, cols = dataset[0][0].size()
    #print(depth, rows, cols)
    n = len(dataset)
    d = torch.zeros((n, depth, rows, cols))
    for i, (image, _) in enumerate(dataset):
        d[i] = image
    c = torch.permute(d, (1, 0, 2, 3))
    mean = torch.zeros(ncolor)
    std  = torch.zeros(ncolor)
    for m in range(ncolor):
        mean[m] = torch.mean(c[m])
        std[m]  = torch.std (c[m])
    return mean, std
